import numpy as np
import pytest

import quimb.tensor as qtn
from quimb.tensor.contraction import _CONTRACT_BACKEND, _TENSOR_LINOP_BACKEND


def test_tensor_contract_strip_exponent():
    tn = qtn.TN_rand_reg(10, 3, 3, dtype=complex)
    z0 = tn.contract()
    m1, e1 = qtn.tensor_contract(*tn, strip_exponent=True)
    assert m1 * 10**e1 == pytest.approx(z0)
    # test tn.exponent is reinserted
    tn.equalize_norms_(value=1.0)
    z2 = tn.contract()
    assert z2 == pytest.approx(z0)
    # test tn.exponent is reinserted with strip exponent
    m3, e3 = tn.contract(strip_exponent=True)
    assert m3 * 10**e3 == pytest.approx(z0)


@pytest.mark.parametrize("strip_exponent", [False, True])
@pytest.mark.parametrize("equalize_norms", [False, 1.0, True])
@pytest.mark.parametrize("inplace", [False, True])
def test_contract_tags_strip_exponent(
    strip_exponent,
    equalize_norms,
    inplace,
):
    tn = qtn.TN_rand_reg(8, 3, 2)
    Zex = tn.contract()

    if inplace:
        tnc = tn.copy()
        tnc.contract_tags_(
            all,
            strip_exponent=strip_exponent,
            equalize_norms=equalize_norms,
        )
        Z = tnc.arrays[0] * 10**tnc.exponent

    else:
        Z = tn.contract_tags(
            all,
            strip_exponent=strip_exponent,
            equalize_norms=equalize_norms,
        )
        if strip_exponent:
            Z = Z[0] * 10 ** Z[1]

    assert Z == pytest.approx(Zex, rel=1e-3)


@pytest.mark.parametrize("strip_exponent", [False, True])
@pytest.mark.parametrize("equalize_norms", [False, 1.0, True])
@pytest.mark.parametrize("inplace", [False, True])
def test_contract_cumulative_strip_exponent(
    strip_exponent,
    equalize_norms,
    inplace,
):
    mps = qtn.MPS_rand_state(7, 3)
    tn = mps.make_norm()
    Zex = tn.contract()

    assert tn._CONTRACT_STRUCTURED

    if inplace:
        tnc = tn.copy()
        tnc.contract_(
            ...,
            strip_exponent=strip_exponent,
            equalize_norms=equalize_norms,
        )
        Z = tnc.arrays[0] * 10**tnc.exponent

    else:
        Z = tn.contract(
            ...,
            strip_exponent=strip_exponent,
            equalize_norms=equalize_norms,
        )
        if strip_exponent:
            Z = Z[0] * 10 ** Z[1]

    assert Z == pytest.approx(Zex, rel=1e-3)


@pytest.mark.parametrize("strip_exponent", [False, True])
@pytest.mark.parametrize("equalize_norms", [False, 1.0, True])
@pytest.mark.parametrize("inplace", [False, True])
def test_contract_compressed_strip_exponent(
    strip_exponent,
    equalize_norms,
    inplace,
):
    L = 6
    tn = qtn.TN2D_rand(L, L, 2, seed=42, dist="uniform")
    Zex = tn.contract()

    if inplace:
        tnc = tn.copy()
        tnc.contract_(
            optimize="greedy-compressed",
            max_bond=4,
            strip_exponent=strip_exponent,
            equalize_norms=equalize_norms,
            progbar=True,
        )
        Z = tnc.arrays[0] * 10**tnc.exponent
    else:
        Z = tn.contract(
            optimize="greedy-compressed",
            max_bond=4,
            strip_exponent=strip_exponent,
            equalize_norms=equalize_norms,
            progbar=True,
        )
        if strip_exponent:
            Z = Z[0] * 10 ** Z[1]

        if equalize_norms == 1.0:
            # undefined
            return

    assert Z == pytest.approx(Zex, rel=1e-3)


class TestContractOpts:
    def test_contract_strategy(self):
        assert qtn.get_contract_strategy() == "greedy"
        with qtn.contract_strategy("auto"):
            assert qtn.get_contract_strategy() == "auto"
        assert qtn.get_contract_strategy() == "greedy"

    def test_contract_backend(self):
        assert qtn.get_contract_backend() == _CONTRACT_BACKEND
        with qtn.contract_backend("cupy"):
            assert qtn.get_contract_backend() == "cupy"
        assert qtn.get_contract_backend() == _CONTRACT_BACKEND

    def test_tensor_linop_backend(self):
        assert qtn.get_tensor_linop_backend() == _TENSOR_LINOP_BACKEND
        with qtn.tensor_linop_backend("cupy"):
            assert qtn.get_tensor_linop_backend() == "cupy"
        assert qtn.get_tensor_linop_backend() == _TENSOR_LINOP_BACKEND

    def test_contract_cache(self):
        import cotengra as ctg

        info = {"num_calls": 0}

        def my_custom_opt(inputs, output, size_dict, memory_limit=None):
            info["num_calls"] += 1
            return [(0, 1)] * (len(inputs) - 1)

        ctg.register_preset("quimb_test_opt", my_custom_opt)

        tn = qtn.MPS_rand_state(4, 3) & qtn.MPS_rand_state(4, 3)
        assert tn.contract(
            all, optimize="quimb_test_opt", get="expression"
        ) is tn.contract(all, optimize="quimb_test_opt", get="expression")
        assert info["num_calls"] == 1

        assert info["num_calls"] == 1


@pytest.mark.parametrize("around", ["I3,3", "I0,0", "I1,2"])
@pytest.mark.parametrize("equalize_norms", [False, True])
@pytest.mark.parametrize("gauge_boundary_only", [False, True])
def test_contract_approx_with_gauges(
    around, equalize_norms, gauge_boundary_only
):
    rng = np.random.default_rng(42)
    tn = qtn.TN2D_from_fill_fn(
        lambda shape: rng.uniform(size=shape, low=-0.5), 7, 7, 4
    )
    Zex = tn ^ ...
    Z = tn.contract_around(
        around,
        max_bond=8,
        gauges=True,
        gauge_boundary_only=gauge_boundary_only,
        tree_gauge_distance=2,
        equalize_norms=equalize_norms,
        progbar=True,
    )
    assert Z == pytest.approx(Zex, rel=1e-2)



"""Helper functions for computing marginals of classical partition functions /
SAT problems.
"""


def compute_all_marginals_via_slicing(
    tn,
    output_inds,
    optimize="auto-hq",
    strip_exponent=False,
    progbar=False,
    **contract_kwargs,
):
    tree0 = tn.contraction_tree(output_inds=(), optimize=optimize)
    arrays = tn.arrays

    w = {}
    Z = None
    symbol_map = tn.get_symbol_map()

    if progbar:
        from tqdm import tqdm

        output_inds = tqdm(output_inds)

    for ix in output_inds:
        # convert quimb index to cotengra unicode symbol
        symbol = symbol_map[ix]
        symbol

        if symbol not in tree0.sliced_inds:
            tree_v = tree0.remove_ind(symbol)
        else:
            # already sliced
            tree_v = tree0

        overall_exponent = None

        # initialize each value of this index
        wv = [0.0 for i in range(tree_v.size_dict[symbol])]

        for s in range(tree_v.nslices):
            # what index combination is this slice
            key = tree_v.slice_key(s)

            if (Z is not None) and key[symbol] == 1:
                # we can compute this from the total count
                continue

            # contract the slice
            if strip_exponent:
                p, exponent = tree_v.contract_slice(
                    arrays, s, **contract_kwargs, strip_exponent=True
                )
            else:
                p = tree_v.contract_slice(arrays, s, **contract_kwargs)
                exponent = 0.0

            p = float(p)
            exponent = float(exponent)

            if overall_exponent is None:
                # set overall exponent from first slice...
                overall_exponent = exponent

            # ... correct subsequent slices relative to first
            rel_exponent = exponent - overall_exponent

            # add result to correct index position
            wv[key[symbol]] += p * 10**rel_exponent

        if Z is None:
            # completed first index, can compute Z
            Z = sum(wv)
            w["Z"] = Z
        else:
            # can use total to infer the 1 value
            wv[1] = Z - wv[0]

        w[ix] = (wv, overall_exponent)

    return w


def compute_all_marginals_via_slicing_shared(
    tn,
    output_inds,
    optimize="auto-hq",
    **contract_kwargs,
):
    from autoray import do, lazy

    tnc = tn.copy()

    with lazy.shared_intermediates():
        tnc.apply_to_arrays(lazy.array)
        w = compute_all_marginals_via_slicing(
            tnc, output_inds=output_inds, optimize=optimize, **contract_kwargs
        )

    # stack into a single lazy array
    return do(
        "stack",
        tuple(
            do("stack", (w[ix][0][0], w[ix][0][1], w[ix][1]))
            for ix in output_inds
        ),
    )


def compute_all_marginals_via_torch_autodiff(
    tn,
    output_inds,
    optimize="auto-hq",
    equalize_norms=1.0,
    contraction_width_error_threshold=float("inf"),
):
    import torch
    from autoray import to_backend_dtype

    tnc = tn.copy()
    if equalize_norms:
        tnc.equalize_norms_(equalize_norms)

    tnc.apply_to_arrays(torch.tensor)

    variables = {}
    for ix in output_inds:
        # create a flat input tensor for each desired output
        data = torch.ones(
            tn.ind_size(ix),
            dtype=to_backend_dtype(tnc.dtype, "torch"),
            requires_grad=True,
        )
        variables[ix] = data
        # multiply it into any connected tensor, so we don't change geometry
        t = next(iter(tnc._inds_get(ix)))
        t.multiply_index_diagonal_(ix, data)

    # find the contraction tree
    tree = tnc.contraction_tree(output_inds=(), optimize=optimize)

    if tree.contraction_width() > contraction_width_error_threshold:
        raise ValueError("Contraction width is above threshold.")

    # perform the forward contraction,
    # compute in log10, mantissa should always be 1
    _, exponent = tree.contract(tnc.arrays, strip_exponent=True)
    exponent.backward()

    # factor is just to correct for log derivative
    return {
        ix: 2.302585092994046 * data.grad.detach().cpu().numpy()
        for ix, data in variables.items()
    }



import quimb.tensor as qtn


def make_w_array(Lx, Ly, i, j, A, B, C, cyclic=False, dtype=None):
    import numpy as np

    try:
        is_cyclic_x, is_cyclic_y = cyclic
    except TypeError:
        is_cyclic_x = is_cyclic_y = cyclic

    if dtype is None:
        dtype = np.common_type(A, B, C)

    # virtual bond dimension
    D = 3
    # physical ind size
    p = 2

    I = np.identity(p, dtype=dtype)
    W = np.zeros((D, D, D, D, p, p), dtype=dtype)

    par = 1
    vac = 0
    ex0 = 2

    base_selector = [vac] * 4 + [slice(None)] * 2

    def select(kv=None, ref=None):
        if ref is None:
            selector = list(base_selector)
        else:
            selector = list(ref)
        if kv is not None:
            for k, v in kv.items():
                r = "urdl".index(k)
                selector[r] = v
        return tuple(selector)

    # shape layout:
    # u  r  d  l  k  b
    # basic vacuum term
    W[select()] = I

    din, dout = {
        # starting corner
        # -> no incoming particle required, branches both up and to the right
        (True, True): ("", "ur"),
        # left boundary
        # -> particle comes from below and branches both up and to the right
        (True, False): ("d", "ur"),
        # bottom boundary
        # -> particle comes from the left and continues to the right
        (False, True): ("l", "r"),
        # bulk
        # -> particle comes from the left and continues to the right
        (False, False): ("l", "r"),
    }[j == 0, i == 0]

    # condition on having incoming particle
    sin = select({d: par for d in din})

    # 1-site decay into vacuum
    # for the starting corner this overwrites the vacuum term
    W[sin] = C

    # basic propagation of particle
    for d in dout:
        W[select({d: par}, sin)] = I

    # start of decay into 2-site interaction, requires input
    for d in "ur":
        W[select({d: ex0}, sin)] = A

    # end of decay into 2-site interaction, into vacuum
    W[select({"d": ex0})] = B
    W[select({"l": ex0})] = B

    # for PBC decay can happen next to particle propagation
    if j == 0 and i > 0:
        W[select({"l": ex0, "d": par, "r": par})] = B

    if i == 0 and j > 0:
        W[select({"d": ex0, "u": par, "l": par})] = B

    if i == 0 and j == 0:
        W[select({"d": ex0, "u": par})] = B
        W[select({"l": ex0, "r": par})] = B

    # project vacuum along boundary
    slicer = []
    for isboundary in [
        # up
        (i == Lx - 1 and not is_cyclic_x),
        # right
        (j == Ly - 1 and not is_cyclic_y),
        # down
        (i == 0 and not is_cyclic_x),
        # left
        (j == 0 and not is_cyclic_y),
    ]:
        slicer.append(vac if isboundary else slice(None))

    return W[tuple(slicer)]


def PEPO_nearest_neighbor(
    A, B, C, Lx, Ly, cyclic=False, dtype=None, **pepo_opts
):
    r"""Create a PEPO for a sum of nearest neighbor interactions:

    .. math::

        H = \sum_{\{i,j\}} A_{i} \otimes B_{j} + \sum_i C_{i}

    where A and B define the left and right hand side of an interaction, and C
    is an on-site term. For example, the transverse field ising model can be
    written with `A = -j * Z`, `B = Z`, and `C = -h * X`.

    Parameters
    ----------
    A : array_like
        The left hand side interaction matrix.
    B : array_like
        The right hand side interaction matrix.
    C : array_like
        The on-site interaction matrix.
    Lx : int
        The number of sites in the x-direction.
    Ly : int
        The number of sites in the y-direction.
    dtype : str or dtype, optional
        The data type of the PEPO tensors. If not provided, it will be inferred
        from A, B, and C.
    pepo_opts : dict, optional
        Additional options to pass to the `qtn.PEPO` constructor.

    Returns
    -------
    PEPO
    """
    Ws = [
        [
            make_w_array(Lx, Ly, i, j, A, B, C, cyclic=cyclic, dtype=dtype)
            for j in range(Ly)
        ]
        for i in range(Lx)
    ]

    return qtn.PEPO(
        Ws,
        shape="urdlkb",
        **pepo_opts,
    )



import autoray as ar
from quimb.tensor import Tensor
from quimb.utils import (
    tree_map,
    tree_unflatten,
    tree_apply,
    Leaf,
)


class ArrayInfo:
    __slots__ = ("shape", "size")

    def __init__(self, shape, size):
        self.shape = shape
        self.size = size

    def __repr__(self):
        return f"<ArrayInfo(shape={self.shape}, size={self.size})>"


class Vectorizer:
    """Object for mapping back and forth between any nested pytree of arrays
    or Tensors and a single flat vector.

    Parameters
    ----------
    tree : pytree of array, optional
        Any nested container of arrays, which will be flattened and packed into
        a single vector.
    """

    def __init__(self, tree=None, backend=None):
        self.infos = None
        self.d = None
        self.ref_tree = None
        self.backend = backend
        self._concatenate = None
        self._reshape = None
        if tree is not None:
            self.setup(tree)

    def setup(self, tree):
        self.infos = []
        self.d = 0

        def extracter(x):
            if isinstance(x, Tensor):
                array = x.data
                size = x.size
                info = x
            else:
                array = x
                shape = ar.do("shape", x, like=self.backend)
                size = ar.do("size", x, like=self.backend)
                info = ArrayInfo(shape, size)

            if self.backend is None:
                # set backend from first array encountered
                self.backend = ar.infer_backend(array)
                self._concatenate = ar.get_lib_fn(self.backend, "concatenate")
                self._reshape = ar.get_lib_fn(self.backend, "reshape")

            self.infos.append(info)
            self.d += size
            return Leaf

        self.ref_tree = tree_map(extracter, tree)

    def pack(self, tree):
        """Take ``arrays`` and pack their values into attribute `.{name}`, by
        default `.vector`.
        """
        if self.infos is None:
            self.setup(tree)

        def extractor(x):
            if isinstance(x, Tensor):
                x = x.data
            arrays.append(self._reshape(x, (-1,)))

        arrays = []
        tree_apply(extractor, tree)
        return self._concatenate(tuple(arrays))

    def unpack(self, vector):
        """Turn the single, flat ``vector`` into a sequence of arrays."""

        def _gen_arrays():
            i = 0
            for info in self.infos:
                # get the linear slice
                f = i + info.size
                array = self._reshape(vector[i:f], info.shape)
                i = f
                if isinstance(info, Tensor):
                    # insert array back into tensor, inplace
                    info.modify(data=array)
                    yield info
                else:
                    yield array

        return tree_unflatten(_gen_arrays(), self.ref_tree)

    def __repr__(self):
        return f"<Vectorizer(d={self.d})>"


class DIIS:
    """Direct Inversion in the Iterative Subspace (DIIS) method (AKA Pulay
    mixing) [1] for converging fixed-point iterations.

    [1] P. Pulay, Convergence acceleration of iterative sequences. The case of
    SCF iteration, 1980, Elsevier, https://doi.org/10.1016/0009-2614(80)80396-4.

    Parameters
    ----------
    max_history : int
        Maximum number of previous guesses to use in extrapolation.
    beta : float
        Mixing parameter, 0.0 means only use input guesses, 1.0 means only use
        extrapolated guesses (original Pulay mixing). Default is 1.0.
    rcond : float
        Cutoff for small singular values in the pseudo-inverse of the B matrix.
        Default is 1e-14.
    """

    def __init__(self, max_history=6, beta=1.0, rcond=1e-14):
        self.max_history = max_history
        self.beta = beta
        self.rcond = rcond

        # storage
        self.vectorizer = Vectorizer()
        self.guesses = [None] * max_history
        self.errors = [None] * max_history
        self.lambdas = []
        self.head = self.max_history - 1

        self.backend = None
        self.B = None
        self.y = None
        self.scalar = None

    def _extrapolate(self):
        # TODO: make this backend agnostic
        import numpy as np

        # XXX: do this all on backend? (though is very small)
        if self.B is None:
            g0 = self.guesses[0]
            self.backend = ar.infer_backend(g0)
            dtype = ar.get_dtype_name(g0)
            self.B = np.zeros((self.max_history + 1,) * 2, dtype=dtype)
            self.y = np.zeros(self.max_history + 1, dtype=dtype)
            self.B[1:, 0] = self.B[0, 1:] = self.y[0] = 1.0
            # define conversion to python scalar
            if "complex" in dtype:
                self.scalar = complex
            else:
                self.scalar = float

        # number of error estimates we have
        d = sum(e is not None for e in self.errors)
        i = self.head
        error_i_conj = self.errors[i].conj()
        for j in range(d):
            cij = self.scalar(error_i_conj @ self.errors[j])
            self.B[i + 1, j + 1] = cij
            if i != j:
                self.B[j + 1, i + 1] = cij.conjugate()

        # solve for coefficients, taking into account rank deficiency
        Binv = np.linalg.pinv(
            self.B[: d + 1, : d + 1],
            rcond=self.rcond,
            hermitian=True,
        )
        coeffs = Binv @ self.y[: d + 1]

        # first entry is -ve. lagrange multiplier -> estimated next residual
        self.lambdas.append(abs(-coeffs[0]))
        coeffs = [self.scalar(c) for c in coeffs[1:]]

        # construct linear combination of previous guesses!
        xnew = ar.do("zeros_like", self.guesses[0], like=self.backend)
        for ci, xi in zip(coeffs, self.guesses):
            xnew += ci * xi

        if self.beta != 0.0:
            # allow custom mix of x + xnew:
            # https://prefetch.eu/know/concept/pulay-mixing/
            # i.e. use not just x_i but also f(x_i) -> y_i
            # original Pulay mixing is beta=1.0 == only xnews
            for ci, ei in zip(coeffs, self.errors):
                xnew += (self.beta * ci) * ei

        return xnew

    def update(self, y):
        """Given new output `y[i]` (the result of `f(x[i])`), update the
        internal state and return the extrapolated next guess `x[i+1]`.

        Parameters
        ----------
        y : pytree of array
            The output of the function `f(x)`. Can be any arbitrary nested
            tree structure with arrays treated at leaves.

        Returns
        -------
        xnext : pytree of array
            The next guess `x[i+1]` to pass to the function `f(x)`, with the
            same tree structure as `y`.
        """
        # convert from pytree -> single real vector
        y = self.vectorizer.pack(y)
        x = self.guesses[self.head]
        if x is None:
            # first guess (no extrapolation)
            xnext = y
        else:
            self.errors[self.head] = y - x
            xnext = self._extrapolate()

        self.head = (self.head + 1) % self.max_history
        # NOTE: copy seems to be necessary here to avoid in-place modifications
        self.guesses[self.head] = ar.do("copy", xnext, like=self.backend)

        # convert new extrapolated guess back to pytree
        return self.vectorizer.unpack(xnext)


class DIISPyscf:
    """Thin wrapper around the PySCF DIIS implementation to handle arbitrary
    pytrees of arrays, for testing purposes."""

    def __init__(self, max_history=6):
        from pyscf.lib.diis import DIIS as PDIIS

        self.pdiis = PDIIS()
        self.pdiis.space = max_history
        self.vectorizer = Vectorizer()

    def update(self, y):
        y = self.vectorizer.pack(y)
        xnext = self.pdiis.update(y)
        return self.vectorizer.unpack(xnext)



"""Tools for 'symbolically' and consitently defining operators and
hamiltonians, which can thenvbe built out into various matrix or tensor network
forms such as sparse matrices and MPOs.
"""

from .builder import (
    SparseOperatorBuilder,
    get_mat,
)
from .hilbertspace import (
    HilbertSpace,
)
from .models import (
    fermi_hubbard_from_edges,
    fermi_hubbard_spinless_from_edges,
    heisenberg_from_edges,
    rand_operator,
)

__all__ = (
    "fermi_hubbard_from_edges",
    "fermi_hubbard_spinless_from_edges",
    "get_mat",
    "heisenberg_from_edges",
    "HilbertSpace",
    "rand_operator",
    "SparseOperatorBuilder",
)
